/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.sensor.fritzmeijer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.EnumMap;
import java.util.List;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.SensorDataFile;
import nl.cloudfarming.client.model.SensorMetadata;
import nl.cloudfarming.client.sensor.SensorEventKey;
import nl.cloudfarming.client.sensor.SensorModule;
import nl.cloudfarming.client.sensor.SensorProcessor;
import nl.cloudfarming.eventbus.BaseGuiEventProducer;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.util.NbBundle;

/**
 * FritzMeijer File processor
 * Processes FritzMeijer sensor files 
 *
 * @author Gerben Feenstra
 */
public class FritzMeijerFileProcessor extends SensorProcessor<FritzMeijerSensorData> {

    // Messages
    private static final String LOGKEY_FAILED_TO_READ_FILE = NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.logerror.failed_to_read_file");
    private static final String LOGKEY_PROCESSING_SUMMARY_ERRORMSG = NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.logerror.nr_of_records_failed"); 
    private static final String LOGKEY_FAILED_TO_CREATE_DATE_FROM_FILENAME = NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.logerror.failed_to_create_date_from_filename");
    private static final String LOGKEY_PROCESSING_START = NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.loginfo.process_start");
    private static final String LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG = NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.loginfo.nr_of_processed");
    private static final String LOGKEY_TYPE_DISPLAYNAME = NbBundle.getMessage(FritzMeijerFileProcessor.class, "sensor.fritzmeijer.fileprocessor.log.displayname");
     
    // Other variable declarations
    private final ProgressHandle progressHandle;
    private static final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("hh:mm:ss");
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("DD.MM.yyyy");
    
    //Tested. mem consumtion diference between 10 and 20 thuosand is 3MB. 20 vs 30 is 50 MB
    private static final int BULK_COMMIT_AMOUNT = 1000;
    
    private static final AppLogger LOG = AppLogFactory.getLogger(FritzMeijerFileProcessor.class);
    private FritzMeijerService service = new FritzMeijerService();
    
    protected File fileToProcess;
    private final EventProducer eventProducer = new EventProducer();
        
    public FritzMeijerFileProcessor(File toProcess) {
        assert toProcess != null;
        this.fileToProcess = toProcess;
        this.progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(SensorModule.class, "sensor.common.fileprocessor.loginfo.process_description", toProcess.getName()));
    }

    /**
     * Parses payload to FritzMeijerSensorData
     * @param payload FritzMeijer sensor line
     * @param dataFile
     * @return FritzMeijerSensorData object
     * @throws ParseException
     * @throws FritzMeijerParseException
     */
    public FritzMeijerSensorData parse(String payload, SensorDataFile dataFile) throws FritzMeijerParseException {
        // First create a entity representation of a file
        //converting the differrent data
        EnumMap map = new EnumMap<FritzMeijerFileDescription, Object>(FritzMeijerFileDescription.class);
        String[] fields = payload.split(FritzMeijerFileDescription.getDelimiter());

        try {
            for (FritzMeijerFileDescription c : FritzMeijerFileDescription.values()) {
                if (c.isMandatory()) {
                    if (c.getTheClass().equals(java.lang.Double.class)) {
                        double theDouble = Double.parseDouble(fields[c.getPosition()]);
                        map.put(c, theDouble);
                    } else if (c.getTheClass().equals(java.lang.Integer.class)) {
                        int theInt = Integer.parseInt(fields[c.getPosition()]);
                        map.put(c, theInt);
                    } else if (c.getTheClass().equals(java.util.Date.class)) {
                        Date theDate = null;
                        if(c.name().equals("DATE")){
                            theDate = DATE_FORMAT.parse(fields[c.getPosition()]);
                        }
                        else if(c.name().equals("TIME")){
                            theDate = TIME_FORMAT.parse(fields[c.getPosition()]);
                        }
                        map.put(c, theDate);
                    } else {
                        map.put(c, fields[c.getPosition()]);
                    }
                }
            }
        } catch (ParseException e) {
            throw new FritzMeijerParseException("Error parsing date ", e);
        } catch (Exception e) {
            throw new FritzMeijerParseException(e);
        }

        FritzMeijerSensorData sensorData = new FritzMeijerSensorData();
        sensorData.setSensorDataFile(dataFile);
        sensorData.setLatitude((Double) map.get(FritzMeijerFileDescription.LATITUDE));
        sensorData.setLongitude((Double) map.get(FritzMeijerFileDescription.LONGITUDE));
        sensorData.setDdate((Date) map.get(FritzMeijerFileDescription.DATE));
        sensorData.setUtcTime((Date) map.get(FritzMeijerFileDescription.TIME));
        sensorData.setCourse((Double) map.get(FritzMeijerFileDescription.HEAD));
        sensorData.setLatitudeOriginal((Double) map.get(FritzMeijerFileDescription.LAT_ORG));
        sensorData.setLongitudeOriginal((Double) map.get(FritzMeijerFileDescription.LON_ORG));
        sensorData.setAltitude((Double) map.get(FritzMeijerFileDescription.ALTITUDE));
        sensorData.setSpeed((Double) map.get(FritzMeijerFileDescription.SPEED));
        sensorData.setFixType((Integer) map.get(FritzMeijerFileDescription.FIXTYPE));
        sensorData.setIrmi((Double) map.get(FritzMeijerFileDescription.IRMI));       
        sensorData.setTcari((Double) map.get(FritzMeijerFileDescription.TCARI));
        sensorData.setOsavi((Double) map.get(FritzMeijerFileDescription.OSAVI));
        sensorData.setTcOs((Double) map.get(FritzMeijerFileDescription.TC_OS));
        sensorData.setWdvi((Double) map.get(FritzMeijerFileDescription.WDVI));
        sensorData.setNdre((Double) map.get(FritzMeijerFileDescription.NDRE));
        return sensorData;
    }
     
    @Override
    public void process() {
        SensorDataFile dataFile = null; 
                
        try {
            int success = 0;
            int progress = 0;
            int fail = 0;
            
            LOG.info(LOGKEY_PROCESSING_START, getLogDisplayName(), fileToProcess.getName());
                     
            // Create DataFile
            dataFile = createSensorDataFile(fileToProcess);
            service.newSensorDataFile(dataFile);

            List<FritzMeijerSensorData> persistBuffer = new ArrayList<FritzMeijerSensorData>();
            
            //First loop through file to determine number of lines. This can be achieved with fileUtils but this will create an out of memory exception with large files.
            BufferedReader br = new BufferedReader(new FileReader(fileToProcess));
            int nrOfLines = 0;
            try {
                while (br.readLine() != null) {
                    nrOfLines++;
                }
            } finally {
                br.close();
            }
            
            try {
                br = new BufferedReader(new FileReader(fileToProcess));

                br.readLine(); // First line (Header) 
                br.readLine(); // Second line (Format)
                br.readLine(); // Third line (Column-types)

                this.progressHandle.start(nrOfLines);
                String line;
                while ((line = br.readLine()) != null) {
                    line = line.trim();
                    if(!line.equals("")){
                        progressHandle.progress(++progress);
                        try {
                            FritzMeijerSensorData toPersist = parse(line, dataFile);
                            persistBuffer.add(toPersist);
                            success++;
                        } catch (Exception ex) {
                            fail++;
                        }

                        if (success % BULK_COMMIT_AMOUNT == 0) {
                            service.bulkPersist(persistBuffer, BULK_COMMIT_AMOUNT);
                            persistBuffer.clear();
                        }
                    }
                }

                LOG.info(LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG, success);
                
                if (fail > 0) {
                    LOG.warning(LOGKEY_PROCESSING_SUMMARY_ERRORMSG, fail);
                }

                //Persist the dataFile. This wil result in persisting the sensordata and metadata as wel.
                if (persistBuffer.size() > 0) {
                    service.bulkPersist(persistBuffer, BULK_COMMIT_AMOUNT);
                }
                eventProducer.triggerEvent(SensorEventKey.SENSOR_DATA_IMPORTED);
                
            } finally {
                progressHandle.finish();
                br.close();
            }
        } catch (ParseException ex) {
            LOG.error(LOGKEY_FAILED_TO_CREATE_DATE_FROM_FILENAME, fileToProcess.getName(), FritzMeijerPreference.SENSOR_DATA_FILE_FORMAT.getValue());
            LOG.trace(ex);
        } catch (Exception ex) {
            LOG.error(LOGKEY_FAILED_TO_READ_FILE, fileToProcess.getName());
            if (dataFile != null) {
                service.removeFailedSensorDataFileAndData(dataFile);
            }
            LOG.trace(ex);
        }
    }
     
    /**
     * Method which will create a SensorMetadata object from the prefferences:
     * FritzMeijerPreference.SENSOR_DATA_FILE_FORMAT
     * FritzMeijerPreference.CROPCIRCLE_SENSOR_HEIGTH
     * FritzMeijerPreference.NUMBER_OF_CROPCIRCLE_SENSORS
     * FritzMeijerPreference.CROPCIRCLE_SENSOR_SPREAD
     * FritzMeijerPreference.SENSOR_TYPE
     * FritzMeijerPreference.SPECTRAL_CHANNELS
     * @return created SensorMetadata object
     */
    @Override
    protected SensorMetadata buildMetadataFromPreferences() {
        String fileFormat = FritzMeijerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
        int sensorHeigth = Integer.parseInt(FritzMeijerPreference.FRITZMEIJER_SENSOR_HEIGTH.getValue());
        int numberOfSensors = Integer.parseInt(FritzMeijerPreference.NUMBER_OF_FRITZMEIJER_SENSORS.getValue());
        String fritzMeijerSensorSpread = FritzMeijerPreference.FRITZMEIJER_SENSOR_SPREAD.getValue();
        String sensorType = FritzMeijerPreference.SENSOR_TYPE.getValue();
        String spectralChannels = FritzMeijerPreference.SPECTRAL_CHANNELS.getValue();
        return new SensorMetadata(fileFormat, sensorHeigth, numberOfSensors, fritzMeijerSensorSpread, sensorType, spectralChannels);
    }

    @Override
    protected String getFileDateFormat() {
        return FritzMeijerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
    }

    @Override
    protected String getLogDisplayName() {
        return LOGKEY_TYPE_DISPLAYNAME;
    }
    
    private class EventProducer extends BaseGuiEventProducer {

        public EventProducer() {
            super(SensorEventKey.class);
        }
    }
}
