/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.sensor.fritzmeijer;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.List;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.sensor.SensorModule;
import nl.cloudfarming.client.sensor.SensorProcessorProvider;
import nl.cloudfarming.client.sensor.SensorUtil;
import nl.cloudfarming.client.util.FilePickerUtil;
import nl.cloudfarming.client.util.MD5Sum;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author Gerben Feenstra
 */
@ServiceProvider(service = SensorProcessorProvider.class)
public class FritzMeijerProcessorProvider implements SensorProcessorProvider {

    private static final AppLogger LOG = AppLogFactory.getLogger(FritzMeijerProcessorProvider.class);
    private static final String LOGKEY_FILE_ALREADY_IMPORTED = NbBundle.getMessage(SensorModule.class, "sensor.common.processorprovider.loginfo.file_already_imported");
    private final String filefilterDescription = NbBundle.getMessage(FritzMeijerProcessorProvider.class, "sensor.fritzmeijer.processorprovider.filefilter.description", FilePickerUtil.getFileFilterTypeString(getSupportedExtentions()));

    @Override
    public List<String> getSupportedExtentions() {
        return Arrays.asList(new String[]{".csv"});
    }

    @Override
    public boolean isSupported(File sensorFile) {
        String fileExtension = sensorFile.getName().substring(sensorFile.getName().lastIndexOf("."));

        // 1: Check extension
        if(!getSupportedExtentions().contains(fileExtension.toLowerCase())){
            return false;
        }

        // 2: Check fileheader
        if(!SensorUtil.hasValidHeader(sensorFile, FritzMeijerFileDescription.values())){
            return false;
        }
        
        // If preceeding code is passed this file is considered supported.
        return true;
    }

    @Override
    public boolean isValid(File sensorFile) {
        try {
            String checkSum = MD5Sum.getHex(MD5Sum.createChecksum(new FileInputStream(sensorFile)));
            String fileName = sensorFile.getName();
            if (new FritzMeijerService().findFileByNameAndChecksum(fileName, checkSum) == null) {
                return true;
            } else {
                LOG.info(LOGKEY_FILE_ALREADY_IMPORTED, fileName);
            }
        } catch (NoSuchAlgorithmException ex) {
            LOG.error(ex.getMessage());
        } catch (IOException ex) {
            LOG.error(ex.getMessage());
        }
        return false;
    }

    @Override
    public String getFormatDescription() {
        return this.filefilterDescription;
    }     
            
    @Override
    public FritzMeijerFileProcessor getProcessor(File sensorFile) {     
       return new FritzMeijerFileProcessor(sensorFile);
    }
}
